import logging
import os
import warnings
from typing import Optional
from urllib.request import urlretrieve

import metatensor.torch as mts
import metatomic.torch as mta
import numpy as np
import torch
import vesin.metatomic as vesin_metatomic
from metatrain.utils.dtype import dtype_to_str
from platformdirs import user_cache_dir
from tqdm import tqdm

from .._models import get_pet_mad
from ._explorer import MADExplorer


warnings.filterwarnings(
    "ignore",
    message=("PET assumes that Cartesian tensors"),
)

VERSIONS = ["latest", "1.0.0"]

PETMAD_EXPLORER_URL = "https://huggingface.co/lab-cosmo/pet-mad-explorer/resolve/{branch}/pet-mad-explorer-latest.ckpt"


METADATA = mta.ModelMetadata(
    name="PET-MAD explorer",
    description="Exploration tool for PET-MAD model features using Sketch-Map projection",  # noqa: E501
    authors=[
        "Arslan Mazitov (arslan.mazitov@epfl.ch)",
        "Sofiia Chorna",
        "Guillaume Fraux",
        "Marnik Bercx",
        "Giovanni Pizzi",
        "Sandip De",
        "Michele Ceriotti (michele.ceriotti@epfl.ch)",
    ],
    references={
        "model": ["http://arxiv.org/abs/2503.14118"],
    },
)


class PETMADFeaturizer:
    """
    Converts structures into low-dimensional projections using PET-MAD features,
    with dimensionality reduction based on sketch-map.

    Usage example:
        >>> import ase.io
        >>> import chemiscope
        >>> from pet_mad.explore import PETMADFeaturizer

        >>> # Load structures
        >>> frames = ase.io.read("dataset.xyz", ":")

        >>> # Create visualization
        >>> chemiscope.explore(frames, featurize=PETMADFeaturizer(version="latest"))
    """

    def __init__(
        self,
        version: str = "latest",
        checkpoint_path: Optional[str] = None,
        pet_checkpoint_path: Optional[str] = None,
        *,
        check_consistency=False,
        device=None,
        length_unit="Angstrom",
        batch_size: int = 1,
        progress_bar=tqdm,
    ):
        """
        :param version: PET-MAD Explorer version to use. Supported versions are
            "latest", "1.0.0".
        :param checkpoint_path: path to a checkpoint file to load the exploration model
            from. If provided, the `version` parameter is ignored.
        :param pet_checkpoint_path: path to a petmad checkpoint file to use for the
            model from. If not provided, the latest checkpoint is fetched from
            HuggingFace.
        :param check_consistency: should we check the model for consistency when
            running, defaults to False.
        :param device: torch device to use for the calculation. If `None`, we will try
            the options in the model's `supported_device` in order.
        :param length_unit: unit of length used in the structures
        """

        if version not in VERSIONS:
            raise ValueError(
                f"Version {version} is not supported. Supported versions are {VERSIONS}"
            )

        cache_dir = user_cache_dir("pet-mad", "metatensor")
        os.makedirs(cache_dir, exist_ok=True)

        petmad_explorer_path = _get_explorer_path(
            version=version,
            checkpoint_path=checkpoint_path,
            cache_dir=cache_dir,
        )

        petmad = get_pet_mad(version="1.1.0", checkpoint_path=pet_checkpoint_path)

        explorer = MADExplorer(petmad.module, device=device)
        explorer.load_checkpoint(petmad_explorer_path)

        outputs = {"features": mta.ModelOutput(per_atom=True)}
        self.dtype = torch.float64

        capabilities = mta.ModelCapabilities(
            outputs=outputs,
            length_unit="angstrom",
            supported_devices=["cpu", "cuda"],
            dtype=dtype_to_str(self.dtype),
            interaction_range=0.0,
            atomic_types=explorer.pet.atomic_types,
        )

        self.mad_explorer = mta.AtomisticModel(explorer.eval(), METADATA, capabilities)

        self.check_consistency = check_consistency
        self.device = device
        self.length_unit = length_unit
        self.batch_size = batch_size
        self.progress_bar = progress_bar

    def __call__(self, frames, environments):
        systems = mta.systems_to_torch(frames)
        vesin_metatomic.compute_requested_neighbors(
            systems,
            self.length_unit,
            self.mad_explorer,
        )

        systems = [s.to(dtype=self.dtype, device=self.device) for s in systems]

        if environments is not None:
            selected_atoms = mts.Labels(
                names=["system", "atom"],
                values=torch.tensor(
                    [(system, atom) for system, atom, _ in environments]
                ),
            )
        else:
            selected_atoms = None

        options = mta.ModelEvaluationOptions(
            length_unit=self.length_unit,
            outputs={"features": mta.ModelOutput(per_atom=True)},
            selected_atoms=selected_atoms,
        )

        outputs = []

        for i_sys in self.progress_bar(range(0, len(systems), self.batch_size)):
            batch_systems = systems[i_sys : i_sys + self.batch_size]
            batch_outputs = self.mad_explorer(
                batch_systems,
                options,
                check_consistency=self.check_consistency,
            )
            outputs.append(
                batch_outputs["features"].block().values.detach().cpu().numpy()
            )

        return np.concatenate(outputs)


def _get_explorer_path(
    version: Optional[str] = None,
    checkpoint_path: Optional[str] = None,
    cache_dir: str = "",
) -> str:
    if checkpoint_path is not None:
        logging.info(f"Loading model from checkpoint: {checkpoint_path}")
        return checkpoint_path

    cached_filename = "pet-mad-explorer-{version}.ckpt".format(version=version)
    branch = f"v{version}" if version != "latest" else "main"
    url = PETMAD_EXPLORER_URL.format(branch=branch)

    cached_path = os.path.join(cache_dir, cached_filename)

    if os.path.exists(cached_path):
        logging.info(f"Loading model from cache: {cached_path}")
        return cached_path

    logging.info("Downloading PET-MAD-Explorer model")
    path, _ = urlretrieve(url, cached_path)

    return path



# PET-MAD
PET_MAD_LATEST_STABLE_VERSION = "1.0.2"
PET_MAD_UQ_AVAILABILITY_VERSION = "1.0.2"
PET_MAD_NC_AVAILABILITY_VERSION = "1.1.0"
PET_MAD_AVAILABLE_VERSIONS = ["1.1.0", "1.0.2", "1.0.1", "1.0.0"]
# PET-MAD DOS
PET_MAD_DOS_LATEST_STABLE_VERSION = "1.0"
PET_MAD_DOS_AVAILABLE_VERSIONS = ["1.0"]



from typing import Optional

import torch


class MLPProjector(torch.nn.Module):
    """
    MLP used to project feature vectors to low-dimensional representations

    :param input_dim: dimensionality of input features
    :param output_dim: target output dimensionality
    """

    def __init__(self, input_dim: int = 1024, output_dim: int = 3):
        super().__init__()

        self.input_dim = input_dim
        self.output_dim = output_dim

        self.fc1 = torch.nn.Linear(self.input_dim, 512)
        self.fc2 = torch.nn.Linear(512, 256)
        self.fc3 = torch.nn.Linear(256, 128)
        self.output = torch.nn.Linear(128, self.output_dim)

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        x = torch.relu(self.fc1(features))
        x = torch.relu(self.fc2(x))
        x = torch.relu(self.fc3(x))
        return self.output(x)


class TorchStandardScaler(torch.nn.Module):
    """The scaler to standatize features by removing the mean and scaling to
    unit variance"""

    def __init__(self, mean: Optional[float] = None, std: Optional[float] = None):
        super().__init__()
        self.register_buffer("mean", mean)
        self.register_buffer("std", std)
        self.eps = 1e-7

    def fit(self, data: torch.Tensor):
        self.mean = data.mean(dim=0)
        self.std = data.std(dim=0)

    def transform(self, data: torch.Tensor) -> torch.Tensor:
        if self.mean is None or self.std is None:
            return data

        return (data - self.mean) / (self.std + self.eps)

    def inverse_transform(self, data: torch.Tensor) -> torch.Tensor:
        if self.mean is None or self.std is None:
            return data

        return data * (self.std + self.eps) + self.mean



import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from urllib.parse import unquote

import numpy as np
import torch
from ase import Atoms
from ase.units import kB
from huggingface_hub import hf_hub_download
from metatomic.torch import ModelMetadata
from scipy.integrate import lebedev_rule
from scipy.spatial.transform import Rotation


hf_pattern = re.compile(
    r"(?P<endpoint>https://[^/]+)/"
    r"(?P<repo_id>[^/]+/[^/]+)/"
    r"resolve/"
    r"(?P<revision>[^/]+)/"
    r"(?P<filename>.+)"
)

NUM_ELECTRONS_PER_ELEMENT = {
    "Al": 3.0,
    "As": 5.0,
    "Ba": 10.0,
    "Be": 4.0,
    "Bi": 15.0,
    "B": 3.0,
    "Br": 7.0,
    "Ca": 10.0,
    "Cd": 12.0,
    "Cl": 7.0,
    "Co": 17.0,
    "C": 4.0,
    "Cr": 14.0,
    "Cs": 9.0,
    "Fe": 16.0,
    "F": 7.0,
    "Ga": 13.0,
    "Ge": 14.0,
    "H": 1.0,
    "In": 13.0,
    "I": 7.0,
    "Ir": 15.0,
    "K": 9.0,
    "Li": 3.0,
    "Mg": 2.0,
    "Mn": 15.0,
    "Na": 9.0,
    "Nb": 13.0,
    "Ni": 18.0,
    "N": 5.0,
    "O": 6.0,
    "Os": 16.0,
    "Pb": 14.0,
    "Po": 16.0,
    "P": 5.0,
    "Pt": 16.0,
    "Re": 15.0,
    "Rn": 18.0,
    "Sb": 15.0,
    "Se": 6.0,
    "Si": 4.0,
    "Sn": 14.0,
    "S": 6.0,
    "Sr": 10.0,
    "Ta": 13.0,
    "Te": 6.0,
    "Ti": 12.0,
    "Tl": 13.0,
    "V": 13.0,
    "W": 14.0,
    "Y": 11.0,
    "Zn": 20.0,
    "Zr": 12.0,
    "Ag": 19.0,
    "Ar": 8.0,
    "Au": 19.0,
    "Ce": 12.0,
    "Dy": 20.0,
    "Er": 22.0,
    "Eu": 17.0,
    "Gd": 18.0,
    "He": 2.0,
    "Hf": 12.0,
    "Hg": 20.0,
    "Ho": 21.0,
    "Kr": 8.0,
    "La": 11.0,
    "Lu": 25.0,
    "Mo": 14.0,
    "Nd": 14.0,
    "Ne": 8.0,
    "Pd": 18.0,
    "Pm": 15.0,
    "Pr": 13.0,
    "Rb": 9.0,
    "Rh": 17.0,
    "Ru": 16.0,
    "Sc": 11.0,
    "Sm": 16.0,
    "Tb": 19.0,
    "Tc": 15.0,
    "Tm": 23.0,
    "Xe": 18.0,
    "Yb": 24.0,
    "Cu": 11.0,
}

AVAILABLE_LEBEDEV_GRID_ORDERS = [
    3,
    5,
    7,
    9,
    11,
    13,
    15,
    17,
    19,
    21,
    23,
    25,
    27,
    29,
    31,
    35,
    41,
    47,
    53,
    59,
    65,
    71,
    77,
    83,
    89,
    95,
    101,
    107,
    113,
    119,
    125,
    131,
]


def get_so3_rotations(
    rotational_average_order: int,
    num_additional_rotations: int,
    axis: Optional[np.ndarray] = None,
) -> List[np.ndarray]:
    axis = np.array([0, 0, 1]) if axis is None else axis
    lebedev_grid = lebedev_rule(rotational_average_order)[0].T

    alphas = np.linspace(0, 2 * np.pi, num_additional_rotations, endpoint=False)
    additional_rotations = [
        Rotation.from_rotvec(axis * alpha).as_matrix() for alpha in alphas
    ]
    lebedev_rotations = [
        Rotation.align_vectors(rot_vector, [0, 0, 1])[0].as_matrix()
        for rot_vector in lebedev_grid
    ]
    rotations: List[np.ndarray] = []
    for lrot in lebedev_rotations:
        for prot in additional_rotations:
            rotations.append(lrot @ prot)
    return rotations


def rotate_atoms(atoms: Atoms, rotations: List[np.ndarray]) -> List[Atoms]:
    rotated_atoms_list = []
    has_cell = atoms.cell is not None and atoms.cell.rank > 0
    for rot in rotations:
        new_atoms = atoms.copy()
        new_atoms.positions = new_atoms.positions @ rot.T
        if has_cell:
            new_atoms.cell = new_atoms.cell @ rot.T
        rotated_atoms_list.append(new_atoms)
    return rotated_atoms_list


def compute_rotational_average(
    results: Dict[str, List[Any]], rotations: List[np.ndarray]
) -> Dict[str, Any]:
    new_results = {}
    for key, value in results.items():
        if "energy" in key:
            new_results[key] = np.mean(value)
            new_results[key + "_rot_std"] = np.std(value)
        elif "forces" in key:
            rotated_back_values = np.array(
                [val @ rot for rot, val in zip(rotations, value, strict=False)]
            )
            new_results[key] = rotated_back_values.mean(axis=0)
            new_results[key + "_rot_std"] = rotated_back_values.std(axis=0)
        elif "stress" in key:
            rotated_back_values = np.array(
                [rot.T @ val @ rot for rot, val in zip(rotations, value, strict=False)]
            )
            new_results[key] = rotated_back_values.mean(axis=0)
            new_results[key + "_rot_std"] = rotated_back_values.std(axis=0)
    return new_results


def get_pet_mad_metadata(version: str):
    return ModelMetadata(
        name=f"PET-MAD v{version}",
        description="A universal interatomic potential for advanced materials modeling",
        authors=[
            "Arslan Mazitov (arslan.mazitov@epfl.ch)",
            "Filippo Bigi",
            "Matthias Kellner",
            "Paolo Pegolo",
            "Davide Tisi",
            "Guillaume Fraux",
            "Sergey Pozdnyakov",
            "Philip Loche",
            "Michele Ceriotti (michele.ceriotti@epfl.ch)",
        ],
        references={
            "architecture": ["https://arxiv.org/abs/2305.19302v3"],
            "model": ["http://arxiv.org/abs/2503.14118"],
        },
    )


def get_pet_mad_dos_metadata(version: str):
    return ModelMetadata(
        name=f"PET-MAD-DOS v{version}",
        description="A universal machine learning model for the electronic density of states",  # noqa: E501
        authors=[
            "Wei Bin How (weibin.how@epfl.ch)",
            "Pol Febrer",
            "Sanggyu Chong",
            "Arslan Mazitov",
            "Filippo Bigi",
            "Matthias Kellner",
            "Sergey Pozdnyakov",
            "Michele Ceriotti (michele.ceriotti@epfl.ch)",
        ],
        references={
            "architecture": ["https://arxiv.org/abs/2508.09000"],
            "model": [],
        },
    )


def fermi_dirac_distribution(
    energies: torch.Tensor, mu: torch.Tensor, T: torch.Tensor
) -> torch.Tensor:
    """
    Fermi-Dirac distribution function.

    :param energies: Energy grid.
    :param mu: Fermi level.
    :param T: Temperature.
    :return: Fermi-Dirac distribution function.
    """
    x = -(energies - mu) / (kB * T)  # Note the negative sign
    return torch.sigmoid(x)


def get_num_electrons(atoms: Union[Atoms, List[Atoms]]) -> torch.Tensor:
    """
    Get the number of electrons for a given ase.Atoms object, or a list of ase.Atoms
    objects.

    :param atoms: ASE atoms object or a list of ASE atoms objects
    :return: Number of electrons for each ase.Atoms object stored in a torch.Tensor
    format.
    """
    num_electrons = []
    if isinstance(atoms, Atoms):
        atoms = [atoms]
    for item in atoms:
        num_electrons.append(
            int(sum([NUM_ELECTRONS_PER_ELEMENT[symbol] for symbol in item.symbols]))
        )
    num_electrons = torch.tensor(num_electrons)
    return num_electrons


def hf_hub_download_url(
    url: str,
    hf_token: Optional[str] = None,
    cache_dir: Optional[Union[str, Path]] = None,
) -> str:
    """Wrapper around `hf_hub_download` allowing passing the URL directly.

    Function is in inverse of `hf_hub_url`
    """

    match = hf_pattern.match(url)

    if not match:
        raise ValueError(f"URL '{url}' has an invalid format for the Hugging Face Hub.")

    endpoint = match.group("endpoint")
    repo_id = match.group("repo_id")
    revision = unquote(match.group("revision"))
    filename = unquote(match.group("filename"))

    # Extract subfolder if applicable
    parts = filename.split("/", 1)
    if len(parts) == 2:
        subfolder, filename = parts
    else:
        subfolder = None
    return hf_hub_download(
        repo_id=repo_id,
        filename=filename,
        subfolder=subfolder,
        cache_dir=cache_dir,
        revision=revision,
        token=hf_token,
        endpoint=endpoint,
    )



import importlib.util
import logging
import warnings
from typing import Optional
from urllib.parse import urlparse
from urllib.request import urlretrieve

import torch
from metatomic.torch import AtomisticModel
from metatrain.utils.io import load_model as load_metatrain_model
from packaging.version import Version

from ._version import (
    PET_MAD_AVAILABLE_VERSIONS,
    PET_MAD_DOS_AVAILABLE_VERSIONS,
    PET_MAD_DOS_LATEST_STABLE_VERSION,
    PET_MAD_LATEST_STABLE_VERSION,
)
from .modules import BandgapModel
from .utils import get_pet_mad_dos_metadata, get_pet_mad_metadata, hf_hub_download_url


BASE_URL = "https://huggingface.co/lab-cosmo/pet-mad/resolve/{tag}/models/pet-mad-{version}.ckpt"


def get_pet_mad(
    *, version: str = "latest", checkpoint_path: Optional[str] = None
) -> AtomisticModel:
    """Get a metatomic ``AtomisticModel`` for PET-MAD.

    :param version: PET-MAD version to use. Defaults to the latest stable version.
    :param checkpoint_path: path to a checkpoint file to load the model from. If
        provided, the `version` parameter is ignored.
    """
    if version == "latest":
        version = Version(PET_MAD_LATEST_STABLE_VERSION)
    if not isinstance(version, Version):
        version = Version(version)

    if version not in [Version(v) for v in PET_MAD_AVAILABLE_VERSIONS]:
        raise ValueError(
            f"Version {version} is not supported. Supported versions are "
            f"{PET_MAD_AVAILABLE_VERSIONS}"
        )

    if version == Version("1.0.0"):
        if not importlib.util.find_spec("pet_neighbors_convert"):
            raise ImportError(
                f"PET-MAD v{version} is now deprecated. Please consider using the "
                "`latest` version. If you still want to use it, please install the "
                "pet-mad package with optional dependencies: "
                "pip install pet-mad[deprecated]"
            )

        import pet_neighbors_convert  # noqa: F401 # pyright: ignore[reportMissingImports]

    if checkpoint_path is not None:
        logging.info(f"Loading PET-MAD model from checkpoint: {checkpoint_path}")
        path = checkpoint_path
    else:
        logging.info(f"Downloading PET-MAD model version: {version}")
        path = BASE_URL.format(tag=f"v{version}", version=f"v{version}")

    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore",
            message="PET assumes that Cartesian tensors of rank 2 are stress-like",
        )
        model = load_metatrain_model(path)

    metadata = get_pet_mad_metadata(version)
    return model.export(metadata)


def save_pet_mad(*, version: str = "latest", checkpoint_path=None, output=None):
    """
    Save the PET-MAD model to a TorchScript file (``pet-mad-xxx.pt``). These files can
    be used with LAMMPS and other tools to run simulations without Python.

    :param version: PET-MAD version to use. Defaults to the latest stable version.
    :param checkpoint_path: path to a checkpoint file to load the model from. If
        provided, the `version` parameter is ignored.
    :param output: path to use for the output model, defaults to
        ``pet-mad-{version}.pt`` when using a version, or the checkpoint path when using
        a checkpoint.
    """
    if version == "latest":
        version = Version(PET_MAD_LATEST_STABLE_VERSION)
    if not isinstance(version, Version):
        version = Version(version)

    extensions_directory = None
    if version == Version("1.0.0"):
        logging.info("Putting TorchScript extensions in `extensions/`")
        extensions_directory = "extensions"

    model = get_pet_mad(version=version, checkpoint_path=checkpoint_path)

    if output is None:
        if checkpoint_path is None:
            output = f"pet-mad-{version}.pt"
        else:
            raise

    model.save(output, collect_extensions=extensions_directory)
    logging.info(f"Saved PET-MAD model to {output}")


BASE_URL_PET_MAD_DOS = "https://huggingface.co/lab-cosmo/pet-mad-dos/resolve/{tag}/models/pet-mad-dos-{version}.pt"
BASE_URL_BANDGAP_MODEL = (
    "https://huggingface.co/lab-cosmo/pet-mad-dos/resolve/{tag}/models/bandgap-model.pt"
)


def get_pet_mad_dos(
    *, version: str = "latest", model_path: Optional[str] = None
) -> AtomisticModel:
    """Get a metatomic ``AtomisticModel`` for PET-MAD-DOS.

    :param version: PET-MAD-DOS version to use. Defaults to latest available version.
    :param model_path: path to a Torch-Scripted metatomic ``AtomisticModel``. If
        provided, the `version` parameter is ignored.
    """
    if version == "latest":
        version = Version(PET_MAD_DOS_LATEST_STABLE_VERSION)
    if not isinstance(version, Version):
        version = Version(version)

    if version not in [Version(v) for v in PET_MAD_DOS_AVAILABLE_VERSIONS]:
        raise ValueError(
            f"Version {version} is not supported. Supported versions are "
            f"{PET_MAD_DOS_AVAILABLE_VERSIONS}"
        )

    if model_path is not None:
        logging.info(f"Loading PET-MAD-DOS model from checkpoint: {model_path}")
        path = model_path
    else:
        logging.info(f"Downloading PET-MAD-DOS model version: {version}")
        path = BASE_URL_PET_MAD_DOS.format(tag=f"v{version}", version=f"v{version}")

    model = load_metatrain_model(path)
    metadata = get_pet_mad_dos_metadata(version)
    model._metadata = metadata
    return model


def _get_bandgap_model(version: str = "latest", model_path: Optional[str] = None):
    """
    Get a bandgap model for PET-MAD-DOS
    """
    if version == "latest":
        version = Version(PET_MAD_DOS_LATEST_STABLE_VERSION)
    if not isinstance(version, Version):
        version = Version(version)

    if version not in [Version(v) for v in PET_MAD_DOS_AVAILABLE_VERSIONS]:
        raise ValueError(
            f"Version {version} is not supported. Supported versions are "
            f"{PET_MAD_DOS_AVAILABLE_VERSIONS}"
        )

    if model_path is not None:
        logging.info(
            f"Loading the PET-MAD-DOS bandgap model from checkpoint: {model_path}"
        )
        path = model_path
    else:
        logging.info(f"Downloading bandgap model version: {version}")
        path = BASE_URL_BANDGAP_MODEL.format(tag=f"v{version}")
        path = str(path)
        url = urlparse(path)

        if url.scheme:
            if url.netloc == "huggingface.co":
                path = hf_hub_download_url(url=url.geturl(), hf_token=None)
            else:
                # Avoid caching generic URLs due to lack of a model hash for proper
                # cache invalidation
                path, _ = urlretrieve(url=url.geturl())

    model = BandgapModel()
    model.load_state_dict(torch.load(path, weights_only=False, map_location="cpu"))
    return model



import torch.nn as nn


class BandgapModel(nn.Module):
    """
    A minimalistic CNN model for predicting bandgap from density of states.
    """

    def __init__(self):
        super(BandgapModel, self).__init__()
        self.conv1 = nn.Conv1d(
            in_channels=1,
            out_channels=64,
            kernel_size=32,
            stride=1,
            dilation=1,
            padding=1,
        )
        self.conv2 = nn.Conv1d(
            in_channels=64, out_channels=64, kernel_size=16, dilation=1, padding=1
        )
        self.conv3 = nn.Conv1d(
            in_channels=64, out_channels=64, kernel_size=8, dilation=1, padding=1
        )
        self.conv4 = nn.Conv1d(
            in_channels=64, out_channels=64, kernel_size=8, dilation=1, padding=1
        )
        self.pool = nn.MaxPool1d(kernel_size=4)
        self.fc1 = nn.Linear(1024, 1024)  # Adjust size after pooling
        self.fc2 = nn.Linear(1024, 1)
        self.silu = nn.SiLU()

    def forward(self, x, last_layer=False):
        x = self.silu(self.conv1(x))  # Conv1 + ReLU
        x = self.pool(x)  # MaxPooling1D
        x = self.silu(self.conv2(x))  # Conv2 + ReLU
        x = self.pool(x)  # MaxPooling1D
        x = self.silu(self.conv3(x))  # Conv2 + ReLU
        x = self.pool(x)  # MaxPooling1D
        x = self.silu(self.conv4(x))  # Conv2 + ReLU
        x = self.pool(x)  # MaxPooling1D
        x = x.view(x.size(0), -1)  # Flatten
        x = self.silu(self.fc1(x))  # Fully Connected Layer 1
        output = self.fc2(x)  # Fully Connected Layer 2 (Output)
        if last_layer:
            return x
        else:
            return output



from pathlib import Path
from typing import Dict, List, Optional, Union

import metatensor.torch as mts
import metatomic.torch as mta
import torch
from metatrain.pet import PET
from metatrain.utils.io import load_model as load_metatrain_model

from ._modules import MLPProjector, TorchStandardScaler


class MADExplorer(torch.nn.Module):
    """
    Metatomic model projecting the last-layer features from PET-MAD into a
    low-dimensional space.

    The model is intended for exploratory analysis and visualization of the learned
    representations.

    :param model: path to a saved PET-MAD checkpoint or an in-memory instance
    :param input_dim: dimensionality of the input PET-MAD features for projector
    :param output_dim: target low dimensionality for the projected embeddings
    :param device: cpu or cuda
    :param features_output: key to access the PET-MAD feature output.
        mtt::aux::energy_last_layer_features is used by default
    """

    def __init__(
        self,
        model: Union[str, Path, PET],
        input_dim: int = 1024,
        output_dim: int = 3,
        device: Optional[Union[str, torch.device]] = "cpu",
        features_output: str = "mtt::aux::energy_last_layer_features",
    ):
        super().__init__()

        self.device = device
        self.features_output = features_output

        if isinstance(model, (str, Path)):
            self.pet = load_metatrain_model(model)
        else:
            self.pet = model

        self.pet = self.pet.to(self.device)
        self.dtype = next(self.pet.parameters()).dtype

        self.projector = MLPProjector(input_dim, output_dim).to(self.device)
        self.feature_scaler = TorchStandardScaler().to(device)
        self.projection_scaler = TorchStandardScaler().to(device)

    def forward(
        self,
        systems: List[mta.System],
        outputs: Dict[str, mta.ModelOutput],
        selected_atoms: Optional[mts.Labels],
    ) -> Dict[str, mts.TensorMap]:
        if list(outputs.keys()) != ["features"]:
            raise ValueError(
                f"`outputs` keys ({', '.join(outputs.keys())}) contain unsupported "
                "keys. Only 'features' is supported"
            )

        systems = [s.to(self.dtype, self.device) for s in systems]

        per_atom = outputs["features"].per_atom
        pet_requested_outputs = {
            self.features_output: mta.ModelOutput(per_atom=per_atom)
        }

        if selected_atoms is not None:
            selected_atoms = selected_atoms.to(self.device)

        features = self._get_features(systems, pet_requested_outputs, selected_atoms)

        if self.feature_scaler.mean is not None and self.feature_scaler.std is not None:
            features = self.feature_scaler.transform(features)

        with torch.no_grad():
            projections = self.projector(features)

        if (
            self.projection_scaler.mean is not None
            and self.projection_scaler.std is not None
        ):
            projections = self.projection_scaler.inverse_transform(projections)

        num_atoms = projections.size(0)
        num_projections = projections.size(1)

        sample_labels = mts.Labels(
            "system", torch.arange(num_atoms, device=self.device).reshape(-1, 1)
        )
        prop_labels = mts.Labels(
            "projection",
            torch.arange(num_projections, device=self.device).reshape(-1, 1),
        )

        if projections.dtype != self.dtype:
            projections = projections.type(self.dtype)

        block = mts.TensorBlock(
            values=projections,
            samples=sample_labels,
            components=[],
            properties=prop_labels,
        )

        tensor_map = mts.TensorMap(
            keys=mts.Labels("_", torch.tensor([[0]], device=self.device)),
            blocks=[block],
        )

        return {"features": tensor_map}

    def _get_features(
        self,
        systems: List[mta.System],
        outputs: Dict[str, mta.ModelOutput],
        selected_atoms: Optional[mts.Labels],
    ) -> torch.Tensor:
        """
        Compute embeddings for the given systems using the PET-MAD model.

        For per-atom features, it concatenates mean and standard deviation of
        features across atoms
        """

        output = self.pet(systems, outputs, selected_atoms)
        features = output[self.features_output]

        if selected_atoms is not None:
            features = mts.slice(features, "samples", selected_atoms)

        if outputs[self.features_output].per_atom:
            mean = mts.mean_over_samples(features, "atom")
            mean_vals = torch.cat([block.values for block in mean.blocks()], dim=0)

            std = mts.std_over_samples(features, "atom")
            std_vals = torch.cat([block.values for block in std.blocks()], dim=0)

            descriptors = torch.cat([mean_vals, std_vals], dim=1)
        else:
            descriptors = features.block().values

        if descriptors.shape[1] != self.projector.input_dim:
            raise ValueError(
                f"Expected input dim for projector: {self.projector.input_dim}, "
                "got: {descriptors.shape[1]}"
            )

        return descriptors.detach()

    def load_checkpoint(self, path: Union[str, Path]):
        checkpoint = torch.load(path, map_location=self.device, weights_only=False)
        self.projector.load_state_dict(checkpoint["projector_state_dict"])

        self.feature_scaler.mean = checkpoint["feature_scaler_mean"]
        self.feature_scaler.std = checkpoint["feature_scaler_std"]
        self.projection_scaler.mean = checkpoint["projection_scaler_mean"]
        self.projection_scaler.std = checkpoint["projection_scaler_std"]



import logging
import os
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from ase import Atoms
from metatomic.torch import ModelOutput
from metatomic.torch.ase_calculator import MetatomicCalculator
from packaging.version import Version
from platformdirs import user_cache_dir

from ._models import _get_bandgap_model, get_pet_mad, get_pet_mad_dos
from ._version import (
    PET_MAD_DOS_LATEST_STABLE_VERSION,
    PET_MAD_LATEST_STABLE_VERSION,
    PET_MAD_NC_AVAILABILITY_VERSION,
    PET_MAD_UQ_AVAILABILITY_VERSION,
)
from .utils import (
    AVAILABLE_LEBEDEV_GRID_ORDERS,
    compute_rotational_average,
    fermi_dirac_distribution,
    get_num_electrons,
    get_so3_rotations,
    rotate_atoms,
)


STR_TO_DTYPE = {
    "float32": torch.float32,
    "float64": torch.float64,
}
DTYPE_TO_STR = {
    torch.float32: "float32",
    torch.float64: "float64",
}


class PETMADCalculator(MetatomicCalculator):
    """
    PET-MAD ASE Calculator
    """

    def __init__(
        self,
        version: str = "latest",
        checkpoint_path: Optional[str] = None,
        calculate_uncertainty: bool = False,
        calculate_ensemble: bool = False,
        rotational_average_order: Optional[int] = None,
        rotational_average_num_additional_rotations: int = 1,
        rotational_average_batch_size: Optional[int] = None,
        dtype: Optional[torch.dtype] = None,
        *,
        check_consistency: bool = False,
        device: Optional[str] = None,
        non_conservative: bool = False,
    ):
        """
        :param version: PET-MAD version to use. Defaults to the latest stable version.
        :param checkpoint_path: path to a checkpoint file to load the model from. If
            provided, the `version` parameter is ignored.
        :param calculate_uncertainty: whether to calculate energy uncertainty.
            Defaults to False. Only available for PET-MAD version 1.0.2.
        :param calculate_ensemble: whether to calculate energy ensemble.
            Defaults to False. Only available for PET-MAD version 1.0.2.
        :param check_consistency: should we check the model for consistency when
            running, defaults to False.
        :param rotational_average_order: order of the Lebedev-Laikov grid used for
            averaging the prediction over rotations.
        :param rotational_average_num_additional_rotations: the number of additional
            rotations sampled from 0 to 2pi angle applied on top of the each
            Lebedev-Laikov rotation vector when performing rotational averaging.
            Defaults to 1, which means that by default only the Lebedev-Laikov grid
            is used for rotational averaging.
        :param rotational_average_batch_size: batch size to use for the rotational
            averaging. If `None`, all rotations will be computed at once.
        :param dtype: dtype to use for the calculations. If `None`, we will use the
            default dtype.
        :param device: torch device to use for the calculation. If `None`, we will try
            the options in the model's `supported_device` in order.
        :param non_conservative: whether to use the non-conservative regime of forces
            and stresses prediction. Defaults to False. Only available for PET-MAD
            version 1.1.0 or higher.

        """

        if version == "latest":
            version = Version(PET_MAD_LATEST_STABLE_VERSION)
        if not isinstance(version, Version):
            version = Version(version)

        if non_conservative and version < Version(PET_MAD_NC_AVAILABILITY_VERSION):
            raise NotImplementedError(
                f"Non-conservative forces and stresses are not available for version "
                f"{version}. Please use PET-MAD version "
                f"{PET_MAD_NC_AVAILABILITY_VERSION} or higher."
            )

        additional_outputs = {}
        if calculate_uncertainty or calculate_ensemble:
            if version < Version(PET_MAD_UQ_AVAILABILITY_VERSION):
                raise NotImplementedError(
                    f"Energy uncertainty and ensemble are not available for version "
                    f"{version}. Please use PET-MAD version "
                    f"{PET_MAD_UQ_AVAILABILITY_VERSION} or higher, or disable the "
                    "calculation of energy uncertainty and energy ensemble."
                )
            else:
                if calculate_uncertainty:
                    additional_outputs["energy_uncertainty"] = ModelOutput(
                        quantity="energy", unit="eV", per_atom=False
                    )
                if calculate_ensemble:
                    additional_outputs["energy_ensemble"] = ModelOutput(
                        quantity="energy", unit="eV", per_atom=False
                    )

        model = get_pet_mad(version=version, checkpoint_path=checkpoint_path)

        if dtype is not None:
            if isinstance(dtype, str):
                assert dtype in STR_TO_DTYPE, f"Invalid dtype: {dtype}"
                dtype = STR_TO_DTYPE[dtype]
            model._capabilities.dtype = DTYPE_TO_STR[dtype]
            model = model.to(dtype=dtype, device=device)

        self._rotations: List[np.ndarray] = []
        if rotational_average_order is not None:
            assert rotational_average_num_additional_rotations > 0, (
                "Number of primitive rotations must be greater than 0."
            )
            if rotational_average_order not in AVAILABLE_LEBEDEV_GRID_ORDERS:
                raise ValueError(
                    f"Lebedev-Laikov grid order {rotational_average_order} is not "
                    f"available. Please use one of the following orders: "
                    f"{AVAILABLE_LEBEDEV_GRID_ORDERS}."
                )

            self._rotations = get_so3_rotations(
                rotational_average_order,
                rotational_average_num_additional_rotations,
            )
        self._rotational_average_batch_size = rotational_average_batch_size

        cache_dir = user_cache_dir("pet-mad", "metatensor")
        os.makedirs(cache_dir, exist_ok=True)

        extensions_directory = None
        if version == Version("1.0.0"):
            extensions_directory = "extensions"

        pt_path = cache_dir + f"/pet-mad-{version}.pt"
        extensions_directory = (
            (cache_dir + "/" + extensions_directory)
            if extensions_directory is not None
            else None
        )

        logging.info(f"Exporting checkpoint to TorchScript at {pt_path}")
        model.save(pt_path, collect_extensions=extensions_directory)

        super().__init__(
            pt_path,
            additional_outputs=additional_outputs,
            extensions_directory=extensions_directory,
            check_consistency=check_consistency,
            device=device,
            non_conservative=non_conservative,
        )

    def calculate(
        self, atoms: Atoms, properties: List[str], system_changes: List[str]
    ) -> None:
        """
        Compute some ``properties`` with this calculator, and return them in the format
        expected by ASE.

        This is not intended to be called directly by users, but to be an implementation
        detail of ``atoms.get_energy()`` and related functions. See
        :py:meth:`ase.calculators.calculator.Calculator.calculate` for more information.

        If the `rotational_average_order` parameter is set during initialization, the
        prediction will be averaged over unique rotations in the Lebedev-Laikov grid of
        a chosen order.

        If the `rotational_average_batch_size` parameter is set during initialization,
        averaging will be performed in batches of the given size to avoid out of memory
        errors.
        """

        super().calculate(atoms, properties, system_changes)

        if len(self._rotations) > 0:
            rotated_atoms_list = rotate_atoms(atoms, self._rotations)
            batch_size = (
                self._rotational_average_batch_size
                if self._rotational_average_batch_size is not None
                else len(rotated_atoms_list)
            )
            batches = [
                rotated_atoms_list[i : i + batch_size]
                for i in range(0, len(rotated_atoms_list), batch_size)
            ]
            results: Dict[str, Any] = {}
            for batch in batches:
                try:
                    batch_results = self.compute_energy(
                        batch, self._do_gradients_with_energy
                    )
                    for key, value in batch_results.items():
                        results.setdefault(key, [])
                        results[key].extend(
                            [value] if isinstance(value, float) else value
                        )
                except torch.cuda.OutOfMemoryError as e:
                    raise RuntimeError(
                        "Out of memory error encountered during rotational averaging. "
                        "Please reduce the batch size or use a lower rotational "
                        "averaging parameters. This can be done by setting the "
                        "`rotational_average_batch_size`, `rotational_average_order`"
                        "and `rotational_average_num_additional_rotations` parameters, "
                        "while initializing the calculator."
                        f"Full error message: {e}"
                    )

            results = compute_rotational_average(results, self._rotations)
            self.results.update(results)

    def _get_uq_output(self, output_name: str):
        if output_name not in self.additional_outputs:
            quantity = output_name.split("_")[1]
            raise ValueError(
                f"Energy {quantity} is not available. Please make sure that you have"
                f" initialized the calculator with `calculate_{quantity}=True` and "
                f"performed evaluation. This option is only available for PET-MAD "
                f"version {PET_MAD_UQ_AVAILABILITY_VERSION} or higher."
            )
        return (
            self.additional_outputs[output_name]
            .block()
            .values.detach()
            .numpy()
            .squeeze()
        )

    def get_energy_uncertainty(self):
        return self._get_uq_output("energy_uncertainty")

    def get_energy_ensemble(self):
        return self._get_uq_output("energy_ensemble")


ENERGY_LOWER_BOUND = -159.6456  # Lower bound of the energy grid for DOS
ENERGY_UPPER_BOUND = 79.1528 + 1.5  # Upper bound of the energy grid for DOS
ENERGY_INTERVAL = 0.05  # Interval of the energy grid for DOS

# If we want to calculate the Fermi level at a given temperature, we need to search
# it around the Fermi level at 0 K. To do this, we first set a certain energy window
# with a certain number of grid points to calculate the integrated DOS. Next, we
# interpolate the integrated DOS to a finer grid and find the Fermi level that
# gives the correct number of electrons.
ENERGY_WINDOW = 0.5
ENERGY_GRID_NUM_POINTS_COARSE = 1000
ENERGY_GRID_NUM_POINTS_FINE = 10000


class PETMADDOSCalculator(MetatomicCalculator):
    """
    PET-MAD DOS Calculator
    """

    def __init__(
        self,
        version: str = "latest",
        model_path: Optional[str] = None,
        bandgap_model_path: Optional[str] = None,
        *,
        check_consistency: bool = False,
        device: Optional[str] = None,
    ):
        """
        :param version: PET-MAD-DOS version to use. Defaults to the latest stable
            version.
        :param model_path: path to a Torch-Scripted model file to load the model from.
            If provided, the `version` parameter is ignored.
        :param bandgap_model_path: path to a PyTorch checkpoint file with the bandgap
            model. If provided, the `version` parameter is ignored.
        :param check_consistency: should we check the model for consistency when
            running, defaults to False.
        :param device: torch device to use for the calculation. If `None`, we will try
            the options in the model's `supported_device` in order.

        """
        if version == "latest":
            version = Version(PET_MAD_DOS_LATEST_STABLE_VERSION)
        if not isinstance(version, Version):
            version = Version(version)

        model = get_pet_mad_dos(version=version, model_path=model_path)
        bandgap_model = _get_bandgap_model(
            version=version, model_path=bandgap_model_path
        )

        super().__init__(
            model,
            additional_outputs={},
            check_consistency=check_consistency,
            device=device,
        )
        self._bandgap_model = bandgap_model

        n_points = np.ceil((ENERGY_UPPER_BOUND - ENERGY_LOWER_BOUND) / ENERGY_INTERVAL)
        self._energy_grid = (
            torch.arange(n_points) * ENERGY_INTERVAL + ENERGY_LOWER_BOUND
        )

    def calculate_dos(
        self, atoms: Union[Atoms, List[Atoms]], per_atom: bool = False
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Calculate the density of states for a given ase.Atoms object,
        or a list of ase.Atoms objects.

        :param atoms: ASE atoms object or a list of ASE atoms objects
        :param per_atom: Whether to return the density of states per atom.
        :return: Energy grid and corresponding DOS values in torch.Tensor format.
        """
        results = self.run_model(
            atoms, outputs={"mtt::dos": ModelOutput(per_atom=per_atom)}
        )
        dos = results["mtt::dos"].block().values
        return self._energy_grid.clone(), dos

    def calculate_bandgap(
        self, atoms: Union[Atoms, List[Atoms]], dos: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Calculate the bandgap for a given ase.Atoms object,or a list of ase.Atoms
        objects. By default, the density of states is first calculated using the
        `calculate_dos` method, and the the bandgap is derived from the DOS by a
        BandgapModel. Alternatively, the density of states can be provided as an
        input parameter to avoid re-calculating the DOS.

        :param atoms: ASE atoms object or a list of ASE atoms objects
        :param dos: Density of states for the given atoms. If not provided, the
            density of states is calculated using the `calculate_dos` method.
        :return: bandgap values for each ase.Atoms object object stored in a
            torch.Tensor format.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        if dos is None:
            _, dos = self.calculate_dos(atoms, per_atom=False)
        if dos.shape[0] != len(atoms):
            raise ValueError(
                f"The provided DOS is inconsistent with the provided `atoms` "
                f"parameter: {len(atoms)} != {dos.shape[0]}. Please either set "
                "`dos = None` or provide a consistent DOS, computed with "
                "`per_atom = False`."
            )
        num_atoms = torch.tensor([len(item) for item in atoms], device=dos.device)
        dos = dos / num_atoms.unsqueeze(1)
        bandgap = self._bandgap_model(
            dos.unsqueeze(1)
        ).detach()  # Need to make the inputs [n_predictions, 1, 4806]
        bandgap = torch.nn.functional.relu(bandgap).squeeze()
        return bandgap

    def calculate_efermi(
        self,
        atoms: Union[Atoms, List[Atoms]],
        dos: Optional[torch.Tensor] = None,
        temperature: float = 0.0,
    ) -> torch.Tensor:
        """
        Get the Fermi energy for a given ase.Atoms object, or a list of ase.Atoms
        objects, based on a predicted density of states at a given temperature.
        By default, the density of states is first calculated using the `calculate_dos`
        method, and the Fermi level is calculated at T=0 K. Alternatively, the density
        of states can be provided as an input parameter to avoid re-calculating the DOS.

        :param atoms: ASE atoms object or a list of ASE atoms objects
        :param dos: Density of states for the given atoms. If not provided, the
            density of states is calculated using the `calculate_dos` method.
        :param temperature: Temperature (K). Defaults to 0 K.
        :return: Fermi energy for each ase.Atoms object stored in a torch.Tensor
        format.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        if dos is None:
            _, dos = self.calculate_dos(atoms, per_atom=False)
        if dos.shape[0] != len(atoms):
            raise ValueError(
                f"The provided DOS is inconsistent with the provided `atoms` "
                f"parameter: {len(atoms)} != {dos.shape[0]}. Please either set "
                "`dos = None` or provide a consistent DOS, computed with "
                "`per_atom = False`."
            )
        cdos = torch.cumulative_trapezoid(dos, dx=ENERGY_INTERVAL)
        num_electrons = get_num_electrons(atoms)
        num_electrons.to(dos.device)
        efermi_indices = torch.argmax(
            (cdos > num_electrons.unsqueeze(1)).float(), dim=1
        )
        efermi = self._energy_grid[efermi_indices]
        if temperature > 0.0:
            efermi_grid_trial = torch.linspace(
                efermi.min() - ENERGY_WINDOW,
                efermi.max() + ENERGY_WINDOW,
                ENERGY_GRID_NUM_POINTS_COARSE,
            )
            occupancies = fermi_dirac_distribution(
                self._energy_grid.unsqueeze(0),
                efermi_grid_trial.unsqueeze(1),
                temperature,
            )
            idos = torch.trapezoid(dos.unsqueeze(1) * occupancies, self._energy_grid)
            idos_interp = torch.nn.functional.interpolate(
                idos.unsqueeze(0),
                size=ENERGY_GRID_NUM_POINTS_FINE,
                mode="linear",
                align_corners=True,
            )[0]
            efermi_grid_interp = torch.nn.functional.interpolate(
                efermi_grid_trial.unsqueeze(0).unsqueeze(0),
                size=ENERGY_GRID_NUM_POINTS_FINE,
                mode="linear",
                align_corners=True,
            )[0][0]
            # Soft approximation of argmax using temperature scaling
            residue = idos_interp - num_electrons.unsqueeze(1)
            # Use softmax with a sharp temperature to approximate argmax
            tau = 0.0001  # Small temperature for sharp approximation
            weights = torch.softmax(-torch.abs(residue) / tau, dim=1)
            efermi = torch.sum(weights * efermi_grid_interp.unsqueeze(0), dim=1)
        return efermi
